from AccessControl.class_init import InitializeClass
from App.special_dtml import DTMLFile
from BTrees.IIBTree import difference
from BTrees.IIBTree import IISet
from logging import getLogger
from OFS.PropertyManager import PropertyManager
from plone.event.recurrence import recurrence_sequence_ical
from plone.event.utils import dt2int
from plone.event.utils import pydt
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PluginIndexes.unindex import UnIndex
from Products.PluginIndexes.util import safe_callable
from ZODB.POSException import ConflictError
from zope.interface import implementer
from zope.interface import Interface
from zope.schema import Text


LOG = getLogger('Products.DateRecurringIndex')
_marker = object()


class IDateRecurringIndex(Interface):
    attr_recurdef = Text(
        title=u"Attribute- or fieldname of recurrence rule definition."
              u"RFC2445 compatible string or timedelta."
    )
    attr_until = Text(
        title=u"Attribute- or fieldname of until date (optional)."
    )


@implementer(IDateRecurringIndex)
class DateRecurringIndex(UnIndex, PropertyManager):
    """Index for dates with recurrence support.
    """

    meta_type = 'DateRecurringIndex'
    query_options = ('query', 'range', 'not')

    manage_main = PageTemplateFile('www/manageDRIndex', globals())
    manage_browse = DTMLFile('www/browseIndex', globals())

    # TODO: for that, this has to be a DTMLFile?
    # manage_main._setName('manage_main')
    manage_options = (
        {'label': 'Settings', 'action': 'manage_main'},
        {'label': 'Browse', 'action': 'manage_browse'},
    ) + PropertyManager.manage_options


    def __init__(self, id, ignore_ex=None, call_methods=None,
                 extra=None, caller=None):
        """ Initialize the index
        @ param extra.recurdef:
        @ param extral.until:
        """
        UnIndex.__init__(self, id, ignore_ex=None, call_methods=None,
                         extra=None, caller=None)
        self.attr_recurdef = extra.recurdef
        self.attr_until = extra.until

    def index_object(self, documentId, obj, threshold=None):
        """index an object, normalizing the indexed value to an integer

           o Normalized value has granularity of one minute.

           o Objects which have 'None' as indexed value are *omitted*,
             by design.

           o Repeat by recurdef - a RFC2445 reccurence definition string

        """
        returnStatus = 0

        try:
            date_attr = getattr(obj, self.id)
            if safe_callable(date_attr):
                date_attr = date_attr()
        except AttributeError:
            return returnStatus

        recurdef = getattr(obj, self.attr_recurdef, None)
        if safe_callable(recurdef):
            recurdef = recurdef()

        if not recurdef:
            dates = [pydt(date_attr)]
        else:
            until = getattr(obj, self.attr_until, None)
            if safe_callable(until):
                until = until()

            dates = recurrence_sequence_ical(
                date_attr, recrule=recurdef, until=until)

        newvalues = IISet(map(dt2int, dates))
        oldvalues = self._unindex.get(documentId, _marker)
        if oldvalues is not _marker:
            oldvalues = IISet(oldvalues)

        if oldvalues is not _marker and newvalues is not _marker\
                and not difference(newvalues, oldvalues)\
                and not difference(oldvalues, newvalues):
            # difference is calculated relative to first argument, so we have
            # to use it twice here
            return returnStatus

        if oldvalues is not _marker:
            for oldvalue in oldvalues:
                self.removeForwardIndexEntry(oldvalue, documentId)
            if newvalues is _marker:
                try:
                    del self._unindex[documentId]
                except ConflictError:
                    raise
                except Exception:
                    LOG.error("Should not happen: oldvalues was there,"
                              " now it's not, for document with id %s" %
                              documentId)

        if newvalues is not _marker:
            inserted = False
            for value in newvalues:
                self.insertForwardIndexEntry(value, documentId)
                inserted = True
            if inserted:
                # store tuple values in reverse index entries for sorting
                self._unindex[documentId] = tuple(newvalues)
                returnStatus = 1

        if returnStatus > 0:
            self._increment_counter()

        return returnStatus

    def unindex_object(self, documentId):
        """ Carefully unindex the object with integer id 'documentId'"""
        values = self._unindex.get(documentId, _marker)
        if values is _marker:
            return None

        for value in values:
            self.removeForwardIndexEntry(value, documentId)

        try:
            del self._unindex[documentId]
        except ConflictError:
            raise
        except Exception:
            LOG.debug('Attempt to unindex nonexistent document'
                      ' with id %s' % documentId, exc_info=True)

    def _convert(self, value, default=None):
        """Convert record keys/datetimes into int representation.
        """
        return dt2int(value) or default


manage_addDRIndexForm = DTMLFile('www/addDRIndex', globals())


def manage_addDRIndex(self, id, extra=None, REQUEST=None, RESPONSE=None,
                      URL3=None):
    """Add a DateRecurringIndex"""
    return self.manage_addIndex(id, 'DateRecurringIndex', extra=extra,
                                REQUEST=REQUEST, RESPONSE=RESPONSE, URL1=URL3)


InitializeClass(DateRecurringIndex)



# -*- coding: utf-8 -*-


def initialize(context):
    # Products.* initialization code is Automatically called by Zope
    from . import index
    context.registerClass(
        index.DateRecurringIndex,
        permission='Add Pluggable Index',
        constructors=(index.manage_addDRIndexForm, index.manage_addDRIndex),
        icon='www/index.gif',
        visibility=None
    )



# -*- coding: utf-8 -*-
"""Installer for the bda.aaf.site package."""

from setuptools import find_packages
from setuptools import setup


version = '3.0.1'
short_description = "Zope 2 date index with support for recurring events."
long_description = ('\n\n'.join([
    open('README.rst').read(),
    open('CONTRIBUTORS.rst').read(),
    open('CHANGES.rst').read(),
    open('LICENSE.rst').read(),
]))


setup(
    name='Products.DateRecurringIndex',
    version=version,
    description=short_description,
    long_description=long_description,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Zope2',
        'Framework :: Zope :: 4',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    keywords='zope zope2 index catalog date recurring',
    author='BlueDynamics Alliance',
    author_email='dev@bluedynamics.com',
    url='https://github.com/collective/Products.DateRecurringIndex',
    license='BSD',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['Products', ],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'BTrees',
        'plone.event',
        'Products.ZCatalog',
        'ZODB',
        'zope.interface',
        'zope.schema',
    ],
    extras_require={
        'test': [
            'pytz',
            'plone.testing'
        ]
    },
)



# -*- coding: utf-8 -*-
import unittest
import doctest

optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS


class DummyEvent(object):
    """some dummy with a start, delta and until to index"""
    def __init__(self, id=None, start=None, recurdef=None, until=None):
        self.id = id
        self.start = start
        self.recurdef = recurdef
        self.until = until


class DummyExtras(object):
    def __init__(self, recurrence_type=None,
                 recurdef=None, until=None):
        self.recurrence_type = recurrence_type
        self.recurdef = recurdef
        self.until = until


class TestIndex(unittest.TestCase):

    def test_index(self):
        """Test the index in icalendar/rfc5545 recurrence mode.
        """

        # Initialize the catalog with DateRecurringIndex
        from Products.DateRecurringIndex.index import DateRecurringIndex

        dri = DateRecurringIndex(
            'start',
            extra=DummyExtras(
                recurrence_type='ical',
                recurdef='recurdef',
                until='until')
        )

        # Index must have be the same name as dri's id
        from Products.ZCatalog.Catalog import Catalog

        cat = Catalog()
        cat.addIndex('start', dri)
        cat.addColumn('id')

        # catalog needs to be contained somewhere, otherwise
        # aquisition-wrapping of result brains doesn't work
        from OFS.Folder import Folder
        portal = Folder(id='portal')
        cat.__parent__ = portal

        # Let's define some dummy events and catalog them.
        from datetime import datetime
        import pytz
        cet = pytz.timezone('CET')

        # Index the same event more than once and test if index size changes.
        test_event = DummyEvent(
            id='test_event',
            start=datetime(2001, 1, 1),
            recurdef='RRULE:FREQ=DAILY;INTERVAL=1;COUNT=5'
        )
        self.assertEqual(
            cat.catalogObject(test_event, 'test_event'),
            1
        )
        self.assertEqual(dri.indexSize(), 5)

        test_event = DummyEvent(
            id='test_event',
            start=datetime(2001, 1, 1),
            recurdef='RRULE:FREQ=DAILY;INTERVAL=1;COUNT=3'
        )
        self.assertEqual(
            cat.catalogObject(test_event, 'test_event'),
            1
        )
        self.assertEqual(dri.indexSize(), 3)

        test_event = DummyEvent(
            id='test_event',
            start=datetime(2001, 1, 1),
            recurdef='RRULE:FREQ=DAILY;INTERVAL=1;COUNT=8'
        )
        self.assertEqual(
            cat.catalogObject(test_event, 'test_event'),
            1
        )
        self.assertEqual(dri.indexSize(), 8)

        cat.uncatalogObject('test_event')
        self.assertEqual(dri.indexSize(), 0)

        # Index for querying later on...
        nonr = DummyEvent(
            id='nonr', start=datetime(2010, 10, 10, 0, 0, tzinfo=cet)
        )
        days = DummyEvent(
            id='days', start=datetime(2010, 10, 10, 0, 0, tzinfo=cet),
            recurdef='RRULE:FREQ=DAILY;INTERVAL=10;COUNT=5'
        )
        mins = DummyEvent(
            id='mins', start=datetime(2010, 10, 10, 0, 0, tzinfo=cet),
            recurdef='RRULE:FREQ=MINUTELY;INTERVAL=10;COUNT=5'
        )
        dstc = DummyEvent(
            id='dstc', start=datetime(2010, 10, 20, 0, 0, tzinfo=cet),
            recurdef='RRULE:FREQ=HOURLY;INTERVAL=1;COUNT=7'
        )

        cat.catalogObject(nonr, 'nonr')
        cat.catalogObject(days, 'days')
        cat.catalogObject(mins, 'mins')
        cat.catalogObject(dstc, 'dstc')

        # Query min one specific date
        query = {
            'start': {
                'query': datetime(2010, 10, 10, 0, 0, tzinfo=cet),
                'range': 'min',
            },
        }
        res = cat(**query)
        self.assertEqual(
            sorted([it.id for it in res]),
            ['days', 'dstc', 'mins', 'nonr']
        )

        # Query max one specific date
        query = {
            'start': {
                'query': datetime(2010, 10, 10, 0, 0, tzinfo=cet),
                'range': 'max',
            },
        }
        res = cat(**query)
        self.assertEqual(
            sorted([it.id for it in res]),
            ['days', 'mins', 'nonr']
        )

        # Query timerange over days and dstc set
        query = {
            'start': {
                'query': [
                    datetime(2010, 10, 11, 0, 0, tzinfo=cet),
                    datetime(2010, 11, 20, 0, 0, tzinfo=cet)
                ],
                'range': 'min:max',
            },
        }
        res = cat(**query)
        self.assertEqual(
            sorted([brain.id for brain in res]),
            ['days', 'dstc']
        )

        # Query timerange over mins set
        query = {
            'start': {
                'query': [
                    datetime(2010, 10, 10, 0, 10, tzinfo=cet),
                    datetime(2010, 10, 10, 0, 40, tzinfo=cet)
                ],
                'range': 'min:max',
            },
        }
        res = cat(**query)
        self.assertEqual(
            sorted([brain.id for brain in res]),
            ['mins']
        )
